WhatsUp = {}

SLASH_WHATSUP1 = "/whatsup"

local _G = _G;

local CAddonID = "B04E9F25743F3A99"

local CHeadingColour = "00F2C300"
local CPlayerDeadColour = "00FF0000"
local CRaidMessagePrefix = "[WhatsUp]:"

local COneHourInSecs = 3600
local CSoulstoneCDInSecs = 1800
local CInnervateCDInSecs = 360
local CRebirthCDInSecs = 1800
local CFearWardDurationInSecs = 600
local CFearWardCDInSecs = 30
local CDivineInterventionDurationInSecs = 180
local CDivineInterventionCDInSecs = 3600

local CDefaultMainformWidth = 160
local CDefaultMainformHeight = 330
local CMinMainformWidth = 160
local CMinMainformHeight = 95
local CMaxMainformWidth = 160
local CMaxMainformHeight = 450

local CMaxRows = 30
local CSoulShardItemID = 6265

local CSoulstoneSpell1ID, CSoulstoneSpell2ID, CSoulstoneSpell3ID, CSoulstoneSpell4ID, CSoulstoneSpell5ID = 20707, 20762, 20763, 20764, 20765
local CInnervateSpellID = 29166
local CRebirthSpell1ID, CRebirthSpell2ID, CRebirthSpell3ID, CRebirthSpell4ID, CRebirthSpell5ID = 20484, 20739, 20742, 20747, 20748
local CFearWardSpellID = 6346
local CDivineInterventionSpellID = 19752

-- Make copies of this for each spell in our declarations
local CSoulstoneSpellLocalisedName = GetSpellInfo(CSoulstoneSpell5ID)
local CInnervateSpellLocalisedName = GetSpellInfo(CInnervateSpellID)
local CRebirthSpellLocalisedName = GetSpellInfo(CRebirthSpell5ID)
local CFearWardSpellLocalisedName = GetSpellInfo(CFearWardSpellID)
local CDivineInterventionSpellLocalisedName = GetSpellInfo(CDivineInterventionSpellID)

local CCommsRequestShardCount = "REQUESTSHARDCOUNT"
local CCommsShardCount = "SHARDCOUNT"
local CCommsSoulstone = "SOULSTONE"
local CCommsFearWard = "FEARWARD"
local CCommsDivineIntervention = "DIVINEINTERVENTION"
local CCommsInnervate = "INNERVATE"
local CCommsRebirth = "REBIRTH"

local SoulstonedList = {}
local SoulstoneCDList = {}
local InnervateCDList = {}
local RebirthCDList = {}
local OptionsList = {}
local FearWardedList = {}
local FearWardCDList = {}
local DivineInterventionList = {}
local DivineInterventionCDList = {}

local RegisterSuccess = false
local IsFrameResizing = false
local SectionToShowFirst = 1
local MaxSections
local IsInARaid = false
local MyInventoryChanged = true
local MyCurrentShardCount = 0

StaticPopupDialogs["WHATSUP_CLEAR_LISTS"] = {
  text = "Do you want to clear this form?",
  button1 = "Yes",
  button2 = "No",
  OnAccept = function() WhatsUp:ClearLists() end,
  timeout = 0,
  whileDead = true,
  hideOnEscape = true,
}

local WhatsUp_EventFrame = CreateFrame("FRAME")

------------------------------------------------------------------------------------
-- Main form functionality ---------------------------------------------------------
------------------------------------------------------------------------------------

function WhatsUp_OnSecond()

	-- Check if we have just joined a raid. Update the guild roster if so.
	if IsInARaid then
		if GetNumGroupMembers() == 0 then
			if OptionsList["ShowHideUponEnteringLeavingGroup"] then
				WhatsUp_MainFrame:Hide()
			end
			IsInARaid = false
		end
		if MyInventoryChanged then
			MyInventoryChanged = false
			local NewShardCount = WhatsUp:GetNumberOfShardsInBags()
			if MyCurrentShardCount ~= NewShardCount then
				MyCurrentShardCount = NewShardCount
				WhatsUp:BroadcastShardCount(MyCurrentShardCount)
			end
		end
	else
		if GetNumGroupMembers() > 0 then
			if OptionsList["ShowHideUponEnteringLeavingGroup"] then
				WhatsUp_MainFrame:Show()
			end
			IsInARaid = true
			WhatsUp:UpdateGuildRoster()
		end
	end

	WhatsUp:UpdateRaidersStatuses()	
	WhatsUp:RedrawPlayerList()
	
end

------------------------------------------------------------------------------------

local function WhatsUp_EventHandler(Self, Event, ...)

	----------------------------------------

	local function CastSoulstone(ASourcePlayerName, ADestPlayerName, ADestPlayerClass)
		WhatsUp:WarlockHasCastSoulstone(ASourcePlayerName)
		WhatsUp:PlayerIsNowSoulstoned(ADestPlayerName, ADestPlayerClass)
	end
	
	local function CastInnervate(ASourcePlayerName)
		WhatsUp:DruidHasCastInnervate(ASourcePlayerName)
	end
	
	local function CastRebirth(ASourcePlayerName)
		WhatsUp:DruidHasCastRebirth(ASourcePlayerName)
	end
	
	local function CastFearWard(ASourcePlayerName, ADestPlayerName, ADestPlayerClass)
		WhatsUp:PriestHasCastFearWard(ASourcePlayerName)
		WhatsUp:PlayerIsNowFearWarded(ADestPlayerName, ADestPlayerClass)
	end
	
	local function CastDivineIntervention(ASourcePlayerName, ADestPlayerName, ADestPlayerClass)
		WhatsUp:PaladinHasCastDivineIntervention(ASourcePlayerName)
		WhatsUp:PlayerIsNowDId(ADestPlayerName, ADestPlayerClass)
	end
	
	----------------------------------------

	if Event == "COMBAT_LOG_EVENT_UNFILTERED" then
		local _, SubEvent, _, _, SourcePlayerName, _, _, DestPlayerGUID, DestPlayerName, _, _, _, SpellName = CombatLogGetCurrentEventInfo()  -- This SpellName appears as whatever the local language is

		if SubEvent == "SPELL_CAST_SUCCESS" then
			if SpellName == CSoulstoneSpellLocalisedName then
				local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayerName)

				if IsInParty then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(DestPlayerName)	
					CastSoulstone(SourcePlayerName, DestPlayerName, DestPlayerClass)

					-- Announce in raid chat only if you are the person casting the spell
					if SourcePlayerName == UnitName("player") then
						WhatsUp:BroadcastMessage(CCommsSoulstone, SourcePlayerName, DestPlayerName)
						
						if OptionsList["AnnounceSpellcastToRaidChat"] then
							SendChatMessage(CRaidMessagePrefix .. " " .. SourcePlayerName .. " has soulstoned " .. DestPlayerName, "RAID", nil, nil)
						end
					end	
				end
			elseif SpellName == CInnervateSpellLocalisedName then
				local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayerName)
			
				if IsInParty then
					CastInnervate(SourcePlayerName)

					-- Announce in raid chat only if you are the person casting the spell
					if SourcePlayerName == UnitName("player") then
						WhatsUp:BroadcastMessage(CCommsInnervate, SourcePlayerName, DestPlayerName)
						
						if OptionsList["AnnounceSpellcastToRaidChat"] then
							SendChatMessage(CRaidMessagePrefix .. " " .. SourcePlayerName .. " has innervated " .. DestPlayerName, "RAID", nil, nil)
						end	
					end		
				end
			elseif SpellName == CRebirthSpellLocalisedName then
				local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayerName)
			
				if IsInParty then
					CastRebirth(SourcePlayerName)

					-- Announce in raid chat only if you are the person casting the spell
					if SourcePlayerName == UnitName("player") then
						WhatsUp:BroadcastMessage(CCommsRebirth, SourcePlayerName, DestPlayerName)
						
						if OptionsList["AnnounceSpellcastToRaidChat"] then
							SendChatMessage(CRaidMessagePrefix .. " " .. SourcePlayerName .. " has battle-rezzed " .. DestPlayerName, "RAID", nil, nil)
						end
					end
				end	
			elseif SpellName == CFearWardSpellLocalisedName then
				local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayerName)
			
				if IsInParty then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(DestPlayerName)	
					CastFearWard(SourcePlayerName, DestPlayerName, DestPlayerClass)
					
					-- Announce in raid chat only if you are the person casting the spell
					if SourcePlayerName == UnitName("player") then
						WhatsUp:BroadcastMessage(CCommsFearWard, SourcePlayerName, DestPlayerName)
						
						if OptionsList["AnnounceSpellcastToRaidChat"] then
							SendChatMessage(CRaidMessagePrefix .. " " .. SourcePlayerName .. " has fear-warded " .. DestPlayerName, "RAID", nil, nil)
						end	
					end	
				end	
			elseif SpellName == CDivineInterventionSpellLocalisedName then
				local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayerName)
			
				if IsInParty then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(DestPlayerName)	
					CastDivineIntervention(SourcePlayerName, DestPlayerName, DestPlayerClass)

					-- Announce in raid chat only if you are the person casting the spell
					if SourcePlayerName == UnitName("player") then
						WhatsUp:BroadcastMessage(CCommsDivineIntervention, SourcePlayerName, DestPlayerName)
						
						if OptionsList["AnnounceSpellcastToRaidChat"] then
							SendChatMessage(CRaidMessagePrefix .. " " .. SourcePlayerName .. " has cast Divine Intervention on " .. DestPlayerName, "RAID", nil, nil)
						end
					end	
				end	
			end
		elseif SubEvent == "SPELL_AURA_REMOVED" then
--			if SpellName == CSoulstoneSpellLocalisedName then
--				WhatsUp:PlayerIsNoLongerSoulstoned(DestPlayerName)
			if SpellName == CFearWardSpellLocalisedName then
				WhatsUp:PlayerIsNoLongerFearWarded(DestPlayerName)
			elseif SpellName == CDivineInterventionSpellLocalisedName then
				WhatsUp:PlayerIsNoLongerDId(DestPlayerName)
			end	
		end
	elseif Event == "CHAT_MSG_ADDON" then
		local AddonID, Message, _, Sender = ...
		local CommsType, SourcePlayer, Word3 = "", "", ""

		Sender = gsub(Sender, "%-[^|]+", "")  --Remove realm name
		
		-- If this raid chat message is meant for this addon...
		if ( AddonID == CAddonID ) and ( Sender ~= UnitName("player") ) then
			for Word in string.gmatch(Message, "%w+") do
				if CommsType == "" then
					CommsType = Word
				elseif SourcePlayer == "" then
					SourcePlayer = Word
				elseif Word3 == "" then
					Word3 = Word
				end
			end
			
			local IsInParty = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(SourcePlayer)			

			if IsInParty then
				if CommsType == CCommsRequestShardCount then
					MyCurrentShardCount = WhatsUp:GetNumberOfShardsInBags()
					WhatsUp:BroadcastShardCount(MyCurrentShardCount)					
				elseif CommsType == CCommsShardCount then
					WhatsUp:SetShardCountForPlayer(SourcePlayer, Word3)
				elseif CommsType == CCommsSoulstone then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(Word3)			
					CastSoulstone(SourcePlayer, Word3, DestPlayerClass)
				elseif CommsType == CCommsInnervate then
					CastInnervate(SourcePlayer)
				elseif CommsType == CCommsRebirth then
					CastRebirth(SourcePlayer)			
				elseif CommsType == CCommsFearWard then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(Word3)			
					CastFearWard(SourcePlayer, Word3, DestPlayerClass)
				elseif CommsType == CCommsDivineIntervention then
					local DestPlayerClass = WhatsUp:GetPlayerClassFromPlayerName(Word3)			
					CastDivineIntervention(SourcePlayer, Word3, DestPlayerClass)
				end
			end
		end
	elseif Event == "GROUP_ROSTER_UPDATE" then
		-- If someone has joined or left the raid, update our lists in case it was someone in them
		WhatsUp:UpdateGuildRoster()
		WhatsUp:UpdateRaidersStatuses()	
		WhatsUp:SortAllLists()
		WhatsUp:RedrawPlayerList()
		WhatsUp:BroadcastMessage(CCommsRequestShardCount, "", "")
	elseif Event == "PLAYER_LOGIN" then
		WhatsUp_MainFrame:SetMinResize(CMinMainformWidth, CMinMainformHeight)	
		WhatsUp_MainFrame:SetMaxResize(CMaxMainformWidth, CMaxMainformHeight)	
		WhatsUp_OptionsFrame:Hide()
		
		-- Load from our global variables
		SoulstonedList = WhatsUp_SoulstonedList
		SoulstoneCDList = WhatsUp_SoulstoneCDList
		InnervateCDList = WhatsUp_InnervateCDList
		RebirthCDList = WhatsUp_RebirthCDList
		OptionsList = WhatsUp_OptionsList	
		FearWardedList = WhatsUp_FearWardedList
		FearWardCDList = WhatsUp_FearWardCDList
		DivineInterventionList = WhatsUp_DivineInterventionList
		DivineInterventionCDList = WhatsUp_DivineInterventionCDList
		
		-- If running for the first time ever, initialise some lists
		if OptionsList == nil then
			OptionsList = {}
		end	
		if OptionsList["IsFormVisible"] == nil then
			OptionsList["IsFormVisible"] = true
		end	
		if OptionsList["LastLogoutTime"] == nil then
			OptionsList["LastLogoutTime"] = 0
		end	
		if OptionsList["AnnounceSpellcastToRaidChat"] == nil then
			OptionsList["AnnounceSpellcastToRaidChat"] = true
		end	
		if OptionsList["IsFormLocked"] == nil then
			OptionsList["IsFormLocked"] = false
		end	
		if OptionsList["BackgroundOpacity"] == nil then
			OptionsList["BackgroundOpacity"] = 0.65
		end	
		if OptionsList["ShowMasterLooter"] == nil then
			OptionsList["ShowMasterLooter"] = true
		end	
		if OptionsList["ShowSoulstonedPlayers"] == nil then
			OptionsList["ShowSoulstonedPlayers"] = true
		end	
		if OptionsList["ShowWarlockSoulstoneCooldowns"] == nil then
			OptionsList["ShowWarlockSoulstoneCooldowns"] = true
		end	
		if OptionsList["ShowDruidInnervateCooldowns"] == nil then
			OptionsList["ShowDruidInnervateCooldowns"] = true
		end	
		if OptionsList["ShowDruidRebirthCooldowns"] == nil then
			OptionsList["ShowDruidRebirthCooldowns"] = true
		end	
		if OptionsList["ShowFearWardedPlayers"] == nil then
			OptionsList["ShowFearWardedPlayers"] = true
		end	
		if OptionsList["ShowPriestFearWardCooldowns"] == nil then
			OptionsList["ShowPriestFearWardCooldowns"] = true
		end
		if OptionsList["ShowDIdPlayers"] == nil then
			OptionsList["ShowDIdPlayers"] = true
		end	
		if OptionsList["ShowPaladinDivineInterventionCooldowns"] == nil then
			OptionsList["ShowPaladinDivineInterventionCooldowns"] = true
		end
		if OptionsList["ShowHideUponEnteringLeavingGroup"] == nil then
			OptionsList["ShowHideUponEnteringLeavingGroup"] = true
		end
		if SoulstonedList == nil then
			SoulstonedList = {}
		end	
		if SoulstoneCDList == nil then
			SoulstoneCDList = {}
		end	
		if InnervateCDList == nil then
			InnervateCDList = {}
		end	
		if RebirthCDList == nil then
			RebirthCDList = {}
		end	
		if FearWardedList == nil then
			FearWardedList = {}
		end	
		if FearWardCDList == nil then
			FearWardCDList = {}
		end	
		if DivineInterventionList == nil then
			DivineInterventionList = {}
		end	
		if DivineInterventionCDList == nil then
			DivineInterventionCDList = {}
		end	
		
		for i = 1,#SoulstoneCDList do
			if SoulstoneCDList[i].SoulShardCount == nil then
				SoulstoneCDList[i].SoulShardCount = 0
			end
		end		
		
		if OptionsList["IsFormVisible"] and RegisterSuccess then
			WhatsUp_MainFrame:Show()
			DEFAULT_CHAT_FRAME:AddMessage("|c00AA00FFWhat's Up (Classic) - Loaded|r")
		elseif RegisterSuccess then
			WhatsUp_MainFrame:Hide()
			DEFAULT_CHAT_FRAME:AddMessage("|c00AA00FFWhat's Up (Classic) - Loaded (type /whatsup to show)|r")
		else	
			DEFAULT_CHAT_FRAME:AddMessage("|c00AA00FFWhat's Up (Classic) - |c00FF0000Registration failed|r")
		end
		
		WhatsUp:ClearOldDataFromLists()
		WhatsUp:SetMaxSections()
		WhatsUp:SetFormLockedIcon()
		WhatsUp:SetBackgroundOpacity()
	elseif Event == "PLAYER_LOGOUT" then
		-- Save back to our global variables
		OptionsList["LastLogoutTime"] = time()
		
		WhatsUp_SoulstonedList = SoulstonedList
		WhatsUp_SoulstoneCDList = SoulstoneCDList
		WhatsUp_InnervateCDList = InnervateCDList
		WhatsUp_RebirthCDList = RebirthCDList
		WhatsUp_OptionsList = OptionsList
		WhatsUp_FearWardedList = FearWardedList
		WhatsUp_FearWardCDList = FearWardCDList
		WhatsUp_DivineInterventionList = DivineInterventionList
		WhatsUp_DivineInterventionCDList = DivineInterventionCDList
	elseif Event == "PLAYER_DEAD" or Event == "PLAYER_UNGHOST" or Event == "PLAYER_ALIVE" then
		WhatsUp:UpdateRaidersStatuses()	
		WhatsUp:RedrawPlayerList()
	elseif Event == "PARTY_LOOT_METHOD_CHANGED" then
		WhatsUp:GetMasterLooter()	
		WhatsUp:RedrawPlayerList()
	elseif Event == "UNIT_INVENTORY_CHANGED" then
		local Arg1 = ...
		if Arg1 == "player" then
			MyInventoryChanged = true
		end
	end

end

------------------------------------------------------------------------------------

function WhatsUp:ClearLists()

	SoulstonedList = {}
	SoulstoneCDList = {}
	InnervateCDList = {}
	RebirthCDList = {}
	FearWardedList = {}
	FearWardCDList = {}
	DivineInterventionList = {}
	DivineInterventionCDList = {}
	
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:ClearOldDataFromLists()

	-- Proceed to reset the old data from the lists, if we were last logged in over an hour ago.
	
	if time() - OptionsList["LastLogoutTime"] > COneHourInSecs then
		WhatsUp:ClearLists()
	end	

end

------------------------------------------------------------------------------------

function WhatsUp:OnMainFormMouseDown()

	if not OptionsList["IsFormLocked"] then
		WhatsUp_MainFrame:StartMoving()
	end	

end

------------------------------------------------------------------------------------

function WhatsUp:OnMainFormMouseUp()

	if not OptionsList["IsFormLocked"] then
		WhatsUp_MainFrame:StopMovingOrSizing()
	end	

end

------------------------------------------------------------------------------------

function WhatsUp:OnOptionsFormMouseDown()

	WhatsUp_OptionsFrame:StartMoving()

end

------------------------------------------------------------------------------------

function WhatsUp:OnOptionsFormMouseUp()

	WhatsUp_OptionsFrame:StopMovingOrSizing()

end

------------------------------------------------------------------------------------

function WhatsUp:OnMainFormMouseWheel(Self, ADelta)

	if ADelta > 0 and SectionToShowFirst > 1 then
		SectionToShowFirst = SectionToShowFirst - 1
	elseif ADelta < 0 and SectionToShowFirst < MaxSections then
		SectionToShowFirst = SectionToShowFirst + 1
	end

	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:OnMainFormShow()

	OptionsList["IsFormVisible"] = true
	
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:OnMainFormHide()

	OptionsList["IsFormVisible"] = false

end

------------------------------------------------------------------------------------

function WhatsUp:OnClearButtonClick()

	StaticPopup_Show("WHATSUP_CLEAR_LISTS")

end

------------------------------------------------------------------------------------

function WhatsUp:OnResizeButtonMouseDown()

	WhatsUp_MainFrame:StartSizing("BOTTOMRIGHT")
	WhatsUp_MainFrame:SetUserPlaced(true)
	
	-- Clear all of the labels
	for i = 1,CMaxRows do
		WhatsUp:RedrawPlayerRowClear(i)
	end	
	
	IsFrameResizing = true

end     

------------------------------------------------------------------------------------

function WhatsUp:OnResizeButtonMouseUp()

	IsFrameResizing = false

	WhatsUp_MainFrame:StopMovingOrSizing()

	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:OnLockButtonClick()

	OptionsList["IsFormLocked"] = not OptionsList["IsFormLocked"]
	WhatsUp:SetFormLockedIcon()

end

------------------------------------------------------------------------------------

function WhatsUp:OnCloseButtonClick()

	WhatsUp_MainFrame:Hide()
	DEFAULT_CHAT_FRAME:AddMessage("|c00AA00FFWhat's Up (Classic) - Hidden (type /whatsup to show)|r")

end

------------------------------------------------------------------------------------

function WhatsUp:SetFormLockedIcon()

	if OptionsList["IsFormLocked"] then
		WhatsUp_LockButton:SetNormalTexture("Interface\\Buttons\\LockButton-Locked-Up")
	else
		WhatsUp_LockButton:SetNormalTexture("Interface\\Buttons\\LockButton-Unlocked-Up")
	end

end

------------------------------------------------------------------------------------

function WhatsUp:SetBackgroundOpacity()

	WhatsUp_MainFrame:SetBackdropColor(1, 1, 1, OptionsList["BackgroundOpacity"])
	WhatsUp_OptionsFrame:SetBackdropColor(1, 1, 1, OptionsList["BackgroundOpacity"])

end

------------------------------------------------------------------------------------

function WhatsUp:GetMasterLooter()

	local _, _, MasterLooterRaidIndex = GetLootMethod()
	
	if MasterLooterRaidIndex == nil then
		OptionsList["MasterLooter"] = nil
	else
		local Name, _, _, _, _, FileName, _, _, _, _, IsMasterLooter = GetRaidRosterInfo(MasterLooterRaidIndex)
		OptionsList["MasterLooter"] = {}
		OptionsList["MasterLooter"].PlayerName = Name
		OptionsList["MasterLooter"].PlayerClass = FileName
	end

end

------------------------------------------------------------------------------------

function WhatsUp:GetPlayerClassFromPlayerName(APlayerName)

	local RaidCount = GetNumGroupMembers()	

	for r = 1, RaidCount do 
		local Name, _, _, _, _, FileName = GetRaidRosterInfo(r)
		if Name == APlayerName then
			return FileName
		end
	end	

	return ""

end

------------------------------------------------------------------------------------

function WhatsUp:UpdateGuildRoster()

	----------------------------------------

	local function AddPlayerToListIfNotThere(APlayerName, AClassName, AList)
	
		local Index = WhatsUp:FindPlayerInList(APlayerName, AList)
		
		if Index == 0 then
			Index = #AList + 1
			
			AList[Index] = {}
			AList[Index].PlayerName = APlayerName
			AList[Index].PlayerClass = AClassName
			AList[Index].CastTime = 0
			AList[Index].SoulShardCount = 0
			AList[Index].IsInParty = true
			AList[Index].IsOnline = true
			AList[Index].IsAlive = true		
		end

		-- If me, then get my shard count and broadcast my shard count
		if ( PlayerName == UnitName("player") ) and ( AClassName == "WARLOCK" ) then
			MyCurrentShardCount = WhatsUp:GetNumberOfShardsInBags()
			WhatsUp:BroadcastShardCount(MyCurrentShardCount)
		end
		
	end

	----------------------------------------

	-- If a player logs on or off, modify the lists of known classes (eg. warlock, druid, dwarf priest).

	local RaidCount = GetNumGroupMembers()	

	for r = 1, RaidCount do 
		local PlayerName, _, _, _, _, FileName = GetRaidRosterInfo(r)
		local _, Race = UnitRace("raid"..r)

		if FileName == "WARLOCK" then
			AddPlayerToListIfNotThere(PlayerName, FileName, SoulstoneCDList)
		elseif FileName == "DRUID" then
			AddPlayerToListIfNotThere(PlayerName, FileName, InnervateCDList)
			AddPlayerToListIfNotThere(PlayerName, FileName, RebirthCDList)
		elseif FileName == "PRIEST" and Race == "Dwarf" then
			AddPlayerToListIfNotThere(PlayerName, FileName, FearWardCDList)
		elseif FileName == "PALADIN" then
			AddPlayerToListIfNotThere(PlayerName, FileName, DivineInterventionCDList)
		end
	end	

end

------------------------------------------------------------------------------------

function WhatsUp:UpdateRaidersStatuses()

	WhatsUp:GetMasterLooter()

	for i,v in pairs(SoulstonedList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
		
		-- If someone on our soulstoned list is no longer alive, and with time remaining on their SS, then mark their SS as active
		if ( not v.IsAlive ) and ( math.floor(time()) - v.CastTime < CSoulstoneCDInSecs ) then
			v.IsSSActive = true
			v.CastTime = 0
		end
		
		-- If we had someone with an active SS, and they're now alive, it means they used the SS, so remove them from the SSed list
		if v.IsSSActive and v.IsAlive then
			v.IsSSActive = false
		end
	end

	for i,v in pairs(SoulstoneCDList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(InnervateCDList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(RebirthCDList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(FearWardedList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(FearWardCDList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(DivineInterventionList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

	for i,v in pairs(DivineInterventionCDList) do
		v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
	end

end	

------------------------------------------------------------------------------------

function WhatsUp:IsPlayerInRaidAndOnlineAndAlive(APlayerName)

	-- Returns if a player is in a raid and is online.
	-- We don't care that if a player is not in the raid, that we return them as offline too. In this case, we really only care if they're in the raid or not.
	-- If a player is in the raid, THEN we care if they're online or not.

	for i = 1,MAX_RAID_MEMBERS do
		local RaiderName, _, _, _, _, _, _, IsRaiderOnline = GetRaidRosterInfo(i)
		
		if APlayerName == RaiderName then
			return true, IsRaiderOnline, not UnitIsDeadOrGhost("raid"..i)
		end
	end

	return false, false, true

end

------------------------------------------------------------------------------------

function WhatsUp:FormatTimeToString(ATimeInSeconds)

	local Minute = math.floor(ATimeInSeconds / 60)
	local Second = ATimeInSeconds % 60
	local TimeText
	
	if Minute > 0 then
		TimeText = Minute .. "m"
	else
		TimeText = ""
	end				
	
	if Second < 10 and Minute > 0 then
		TimeText = TimeText .. "0" .. Second .. "s"
	else	
		TimeText = TimeText .. Second .. "s"
	end

	return TimeText

end

------------------------------------------------------------------------------------

function WhatsUp:RedrawPlayerRow(ALineNumber, AText)

	if ALineNumber <= CMaxRows then
		local CurrentRowBottom = _G["WhatsUp_PlayerName"..ALineNumber]:GetTop() - _G["WhatsUp_PlayerName"..ALineNumber]:GetHeight()

		-- Don't draw the row if it's off-screen
		if CurrentRowBottom - WhatsUp_ClearButton:GetTop() > 15 then
			_G["WhatsUp_PlayerName" .. ALineNumber]:SetText(AText)
		end
	end

	return ALineNumber + 1

end

------------------------------------------------------------------------------------

function WhatsUp:RedrawPlayerRowNone(ALineNumber)

	local LineNumber = WhatsUp:RedrawPlayerRow(ALineNumber, "|c00BBBBBB   < None >")

	return LineNumber

end

------------------------------------------------------------------------------------

function WhatsUp:RedrawPlayerRowClear(ALineNumber)

	-- Clear the row regardless of where it is positioned

	_G["WhatsUp_PlayerName" .. ALineNumber]:SetText("")

end

------------------------------------------------------------------------------------

function WhatsUp:RedrawPlayerList()

	if not OptionsList["IsFormVisible"] then
		return
	end
	
	if IsFrameResizing then
		return
	end

	local LineNumber, CurrentSection, HexColour, CurrentTime, TimeRemaining, Minute, Second
	local AreMasterLooterVisible, AreSoulstonedVisible, AreSoulstoneCDVisible, AreFearWardedVisible, AreFearWardCDVisible = false, false, false, false, false
	local AreDIdVisible, AreDivineInterventionCDVisible, AreInnervateCDVisible, AreRebirthCDVisible = false, false, false, false
	
	LineNumber = 1
	CurrentSection = 0
	CurrentTime = math.floor(time())
	
	-- Clear all of the labels
	for i = 1,CMaxRows do
		WhatsUp:RedrawPlayerRowClear(i)
	end		


	-- Draw the master looter
	if OptionsList["ShowMasterLooter"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Master Looter")
			if OptionsList["MasterLooter"] ~= nil then
				AreMasterLooterVisible = true
				_,_,_,HexColour = GetClassColor(OptionsList["MasterLooter"].PlayerClass)
					
				LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. OptionsList["MasterLooter"].PlayerName)
			end
		
			if not AreMasterLooterVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end
	end


	-- Draw all those who are currently soulstoned
	if OptionsList["ShowSoulstonedPlayers"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Soulstoned")
			local TimeMsg
			for i,v in pairs(SoulstonedList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CSoulstoneCDInSecs - (CurrentTime - v.CastTime)

					if ( TimeRemaining > 0 ) or v.IsSSActive then
						AreSoulstonedVisible = true
						TimeMsg = ""
						
						-- Check if a player has returned to life without us knowing
						if not v.IsAlive then
							v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
						end
						
						if v.IsAlive then
							_,_,_,HexColour = GetClassColor(v.PlayerClass)
							TimeMsg = WhatsUp:FormatTimeToString(TimeRemaining)
						else
							HexColour = CPlayerDeadColour
							TimeMsg = "SS Active"
						end
						
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00FFFFFF - " .. TimeMsg)
					end
				end
			end
		
			if not AreSoulstonedVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end
	end


	-- Draw all our warlocks
	if OptionsList["ShowWarlockSoulstoneCooldowns"] then
		CurrentSection = CurrentSection + 1
		local ShardCount = ""

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Soulstone CD")
			for i,v in pairs(SoulstoneCDList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CSoulstoneCDInSecs - (CurrentTime - v.CastTime)
					AreSoulstoneCDVisible = true

					-- Check if a player has returned to life without us knowing
					if not v.IsAlive then
						v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
					end

					if v.IsAlive then
						_,_,_,HexColour = GetClassColor(v.PlayerClass)
					else
						HexColour = CPlayerDeadColour
					end
					ShardCount = "|c00FFFFFF (" .. v.SoulShardCount .. ") "
					
					if TimeRemaining > 0 then
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. ShardCount .. "|c00BBBBBB - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					else
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. ShardCount .. "|c00BBBBBB - " .. "|c0000BB00Free")
					end
				end
			end

			if not AreSoulstoneCDVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end	
	end	
	
	
	-- Draw all those who are currently fear-warded
	if OptionsList["ShowFearWardedPlayers"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Fear-Warded")
			for i,v in pairs(FearWardedList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CFearWardDurationInSecs - (CurrentTime - v.CastTime)

					if TimeRemaining > 0 then
						AreFearWardedVisible = true
						
						-- Check if a player has returned to life without us knowing
						if not v.IsAlive then
							v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
						end
						
						if v.IsAlive then
							_,_,_,HexColour = GetClassColor(v.PlayerClass)
						else
							HexColour = CPlayerDeadColour
						end

						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00FFFFFF - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					end
				end
			end
		
			if not AreFearWardedVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end
	end


	-- Draw all our dwarf priests
	if OptionsList["ShowPriestFearWardCooldowns"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Fear Ward CD")
			
			for i,v in pairs(FearWardCDList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CFearWardCDInSecs - (CurrentTime - v.CastTime)
					AreFearWardCDVisible = true
					
					-- Check if a player has returned to life without us knowing
					if not v.IsAlive then
						v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
					end
					
					if v.IsAlive then
						_,_,_,HexColour = GetClassColor(v.PlayerClass)
					else
						HexColour = CPlayerDeadColour
					end
						
					if TimeRemaining > 0 then
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					else
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. "|c0000BB00Free")
					end
				end
			end

			if not AreFearWardCDVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end	
	end	
	
	
	-- Draw all those who are currently DI'd
	if OptionsList["ShowDIdPlayers"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Divine Interventioned")
			for i,v in pairs(DivineInterventionList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CDivineInterventionDurationInSecs - (CurrentTime - v.CastTime)

					if TimeRemaining > 0 then
						AreDIdVisible = true
						
						-- Check if a player has returned to life without us knowing
						if not v.IsAlive then
							v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
						end
						
						if v.IsAlive then
							_,_,_,HexColour = GetClassColor(v.PlayerClass)
						else
							HexColour = CPlayerDeadColour
						end

						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00FFFFFF - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					end
				end
			end
		
			if not AreDIdVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end
	end


	-- Draw all our paladins
	if OptionsList["ShowPaladinDivineInterventionCooldowns"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Divine Intervention CD")
			
			for i,v in pairs(DivineInterventionCDList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CDivineInterventionCDInSecs - (CurrentTime - v.CastTime)
					AreDivineInterventionCDVisible = true
					
					-- Check if a player has returned to life without us knowing
					if not v.IsAlive then
						v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
					end
					
					if v.IsAlive then
						_,_,_,HexColour = GetClassColor(v.PlayerClass)
					else
						HexColour = CPlayerDeadColour
					end
						
					if TimeRemaining > 0 then
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					else
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. "|c0000BB00Free")
					end
				end
			end

			if not AreDivineInterventionCDVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end	
	end	
	

	-- Draw all our druids
	if OptionsList["ShowDruidInnervateCooldowns"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Innervate CD")
			
			for i,v in pairs(InnervateCDList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CInnervateCDInSecs - (CurrentTime - v.CastTime)
					AreInnervateCDVisible = true
					
					-- Check if a player has returned to life without us knowing
					if not v.IsAlive then
						v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
					end
					
					if v.IsAlive then
						_,_,_,HexColour = GetClassColor(v.PlayerClass)
					else
						HexColour = CPlayerDeadColour
					end

					if TimeRemaining > 0 then
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					else
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. "|c0000BB00Free")
					end
				end
			end
		
			if not AreInnervateCDVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end
	end
	
	
	if OptionsList["ShowDruidRebirthCooldowns"] then
		CurrentSection = CurrentSection + 1

		if SectionToShowFirst <= CurrentSection then
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. CHeadingColour .. "Rebirth CD")
			
			for i,v in pairs(RebirthCDList) do
				if v.IsInParty and v.IsOnline then	
					TimeRemaining = CRebirthCDInSecs - (CurrentTime - v.CastTime)
					AreRebirthCDVisible = true
					
					-- Check if a player has returned to life without us knowing
					if not v.IsAlive then
						v.IsInParty, v.IsOnline, v.IsAlive = WhatsUp:IsPlayerInRaidAndOnlineAndAlive(v.PlayerName)
					end
					
					if v.IsAlive then
						_,_,_,HexColour = GetClassColor(v.PlayerClass)
					else
						HexColour = CPlayerDeadColour
					end

					if TimeRemaining > 0 then
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. WhatsUp:FormatTimeToString(TimeRemaining))
					else
						LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "|c" .. HexColour .. "   " .. v.PlayerName .. "|c00BBBBBB - " .. "|c0000BB00Free")
					end
				end
			end
			
			if not AreRebirthCDVisible then
				LineNumber = WhatsUp:RedrawPlayerRowNone(LineNumber)
			end	
			LineNumber = WhatsUp:RedrawPlayerRow(LineNumber, "")
		end	
	end

end

------------------------------------------------------------------------------------

function WhatsUp:FindPlayerInList(ASourcePlayerName, AList)

	-- Find the player, if they already exist in our list

	for i = 1,#AList do
		if ASourcePlayerName == AList[i].PlayerName then
			return i
		end
	end
	
	return 0

end

------------------------------------------------------------------------------------

function WhatsUp:SortList(AList)

	table.sort(AList,	function(a,b) 
							if a.CastTime == b.CastTime then
								return a.PlayerName < b.PlayerName
							else	
								return a.CastTime < b.CastTime
							end	
						end)
	
end

------------------------------------------------------------------------------------

function WhatsUp:SortAllLists()

	WhatsUp:SortList(SoulstoneCDList)
	WhatsUp:SortList(SoulstonedList)
	WhatsUp:SortList(InnervateCDList)
	WhatsUp:SortList(RebirthCDList)
	WhatsUp:SortList(FearWardCDList)
	WhatsUp:SortList(FearWardedList)
	WhatsUp:SortList(DivineInterventionCDList)
	WhatsUp:SortList(DivineInterventionList)

end

------------------------------------------------------------------------------------

function WhatsUp:WarlockHasCastSoulstone(ASourcePlayerName)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, SoulstoneCDList)	

--	if Index == 0 then
--		Index = #SoulstoneCDList + 1
--	end
	
--	SoulstoneCDList[Index] = {}
--	SoulstoneCDList[Index].PlayerName = ASourcePlayerName
--	SoulstoneCDList[Index].PlayerClass = "WARLOCK"
	SoulstoneCDList[Index].CastTime = math.floor(time())
--	if SoulstoneCDList[Index].PlayerName == UnitName("player") then
--		SoulstoneCDList[Index].SoulShardCount = WhatsUp:GetNumberOfShardsInBags()
--		WhatsUp:BroadcastShardCount(SoulstoneCDList[Index].SoulShardCount)
--	end	
--	SoulstoneCDList[Index].IsInParty = true
--	SoulstoneCDList[Index].IsOnline = true
--	SoulstoneCDList[Index].IsAlive = true
	
	WhatsUp:SortList(SoulstoneCDList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PlayerIsNowSoulstoned(ASourcePlayerName, APlayerClass)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, SoulstonedList)	

	if Index == 0 then
		Index = #SoulstonedList + 1
	end
	
	SoulstonedList[Index] = {}
	SoulstonedList[Index].PlayerName = ASourcePlayerName
	SoulstonedList[Index].PlayerClass = APlayerClass
	SoulstonedList[Index].CastTime = math.floor(time())
	SoulstonedList[Index].IsInParty = true
	SoulstonedList[Index].IsOnline = true
	SoulstonedList[Index].IsAlive = true
	SoulstonedList[Index].IsSSActive = false
	
	WhatsUp:SortList(SoulstonedList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:DruidHasCastInnervate(ASourcePlayerName)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, InnervateCDList)	

	if Index == 0 then
		Index = #InnervateCDList + 1
	end
		
	InnervateCDList[Index] = {}
	InnervateCDList[Index].PlayerName = ASourcePlayerName
	InnervateCDList[Index].PlayerClass = "DRUID"
	InnervateCDList[Index].CastTime = math.floor(time())
	InnervateCDList[Index].IsInParty = true
	InnervateCDList[Index].IsOnline = true
	InnervateCDList[Index].IsAlive = true
	
	WhatsUp:SortList(InnervateCDList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:DruidHasCastRebirth(ASourcePlayerName)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, RebirthCDList)	

	if Index == 0 then
		Index = #RebirthCDList + 1
	end
		
	RebirthCDList[Index] = {}
	RebirthCDList[Index].PlayerName = ASourcePlayerName
	RebirthCDList[Index].PlayerClass = "DRUID"
	RebirthCDList[Index].CastTime = math.floor(time())
	RebirthCDList[Index].IsInParty = true
	RebirthCDList[Index].IsOnline = true
	RebirthCDList[Index].IsAlive = true
	
	WhatsUp:SortList(RebirthCDList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PriestHasCastFearWard(ASourcePlayerName)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, FearWardCDList)	
	
	if Index == 0 then
		Index = #FearWardCDList + 1
	end
		
	FearWardCDList[Index] = {}
	FearWardCDList[Index].PlayerName = ASourcePlayerName
	FearWardCDList[Index].PlayerClass = "PRIEST"
	FearWardCDList[Index].CastTime = math.floor(time())
	FearWardCDList[Index].IsInParty = true
	FearWardCDList[Index].IsOnline = true
	FearWardCDList[Index].IsAlive = true
	
	WhatsUp:SortList(FearWardCDList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PlayerIsNowFearWarded(ASourcePlayerName, APlayerClass)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, FearWardedList)	

	if Index == 0 then
		Index = #FearWardedList + 1
	end
	
	FearWardedList[Index] = {}
	FearWardedList[Index].PlayerName = ASourcePlayerName
	FearWardedList[Index].PlayerClass = APlayerClass
	FearWardedList[Index].CastTime = math.floor(time())
	FearWardedList[Index].IsInParty = true
	FearWardedList[Index].IsOnline = true
	FearWardedList[Index].IsAlive = true
	
	WhatsUp:SortList(FearWardedList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PaladinHasCastDivineIntervention(ASourcePlayerName)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, DivineInterventionCDList)	
	
	if Index == 0 then
		Index = #DivineInterventionCDList + 1
	end
		
	DivineInterventionCDList[Index] = {}
	DivineInterventionCDList[Index].PlayerName = ASourcePlayerName
	DivineInterventionCDList[Index].PlayerClass = "PALADIN"
	DivineInterventionCDList[Index].CastTime = math.floor(time())
	DivineInterventionCDList[Index].IsInParty = true
	DivineInterventionCDList[Index].IsOnline = true
	DivineInterventionCDList[Index].IsAlive = true
	
	WhatsUp:SortList(DivineInterventionCDList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PlayerIsNowDId(ASourcePlayerName, APlayerClass)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayerName, DivineInterventionList)	

	if Index == 0 then
		Index = #DivineInterventionList + 1
	end
	
	DivineInterventionList[Index] = {}
	DivineInterventionList[Index].PlayerName = ASourcePlayerName
	DivineInterventionList[Index].PlayerClass = APlayerClass
	DivineInterventionList[Index].CastTime = math.floor(time())
	DivineInterventionList[Index].IsInParty = true
	DivineInterventionList[Index].IsOnline = true
	DivineInterventionList[Index].IsAlive = true
	
	WhatsUp:SortList(DivineInterventionList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------
--[[
function WhatsUp:PlayerIsNoLongerSoulstoned(APlayerName)

	-- Either the player has died, or has manually removed their SS.

	for i,v in pairs(SoulstonedList) do
		if v.PlayerName == APlayerName then
			v.CastTime = 0
			break
		end
	end
	
	WhatsUp:SortList(SoulstonedList)
	WhatsUp:RedrawPlayerList()

end
--]]
------------------------------------------------------------------------------------

function WhatsUp:PlayerIsNoLongerFearWarded(APlayerName)

	for i,v in pairs(FearWardedList) do
		if v.PlayerName == APlayerName then
			v.CastTime = 0
			break
		end
	end
	
	WhatsUp:SortList(FearWardedList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:PlayerIsNoLongerDId(APlayerName)

	for i,v in pairs(DivineInterventionList) do
		if v.PlayerName == APlayerName then
			v.CastTime = 0
			break
		end
	end
	
	WhatsUp:SortList(DivineInterventionList)
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:GetNumberOfShardsInBags()

	local BagNo, SlotNo
	local ShardCount = 0
 
	for BagNo = 0, NUM_BAG_SLOTS do
		for SlotNo = 1, GetContainerNumSlots(BagNo) do
			if GetContainerItemID(BagNo, SlotNo) == CSoulShardItemID then
				ShardCount = ShardCount + 1
			end
		end
	end
	
	WhatsUp:SetShardCountForPlayer(UnitName("player"), ShardCount)
	
	return ShardCount

end

------------------------------------------------------------------------------------

function WhatsUp:SetShardCountForPlayer(ASourcePlayer, AShardCount)

	local Index = WhatsUp:FindPlayerInList(ASourcePlayer, SoulstoneCDList)
	
	if Index ~= 0 then
		SoulstoneCDList[Index].SoulShardCount = AShardCount	
	end

end

------------------------------------------------------------------------------------

function WhatsUp:BroadcastShardCount(AShardCount)

	C_ChatInfo.SendAddonMessage( CAddonID, CCommsShardCount .. " " .. UnitName("player") .. " " .. AShardCount, "RAID" )

end

------------------------------------------------------------------------------------

function WhatsUp:BroadcastMessage(AMessageType, ASourcePlayer, ADestPlayer)

	C_ChatInfo.SendAddonMessage( CAddonID, AMessageType .. " " .. ASourcePlayer .. " " .. ADestPlayer, "RAID" )

end

------------------------------------------------------------------------------------

function SlashCmdList.WHATSUP(AMsg, AEditBox)

	AMsg = string.lower(AMsg)
	
	if AMsg == "" or AMsg == "show" then
		WhatsUp_MainFrame:Show()
		OptionsList["IsFormVisible"] = true
	elseif AMsg == "reset" then
		WhatsUp_MainFrame:SetSize(CDefaultMainformWidth, CDefaultMainformHeight)
		WhatsUp_MainFrame:ClearAllPoints()
		WhatsUp_MainFrame:SetPoint("CENTER", 0, 0)

		WhatsUp_OptionsFrame:ClearAllPoints()
		WhatsUp_OptionsFrame:SetPoint("CENTER", 0, 0)

		OptionsList["IsFormLocked"] = false
		WhatsUp:SetFormLockedIcon()
		WhatsUp_MainFrame:Show()
	end	

end	

------------------------------------------------------------------------------------
-- Options form functionality ------------------------------------------------------
------------------------------------------------------------------------------------

function WhatsUp:OnOpenOptionsButtonClick()

	-- Set all UI options prior to opening the options form
	
	WhatsUp_MasterLooterCheckboxText:SetText("Show Master Looter")
	WhatsUp_ShowSoulstonedPlayersCheckboxText:SetText("Show Soulstoned Players")
	WhatsUp_ShowWarlockSoulstoneCooldownsCheckboxText:SetText("Show Warlock Soulstone Cooldowns")
	WhatsUp_ShowFearWardedPlayersCheckboxText:SetText("Show Fear-Warded Players")
	WhatsUp_ShowPriestFearWardCooldownsCheckboxText:SetText("Show Priest Fear Ward Cooldowns")
	WhatsUp_ShowDIdPlayersCheckboxText:SetText("Show Divine Interventioned Players")
	WhatsUp_ShowPaladinDivineInterventionCooldownsCheckboxText:SetText("Show Paladin Divine Intervention Cooldowns")
	WhatsUp_ShowDruidInnervateCooldownsCheckboxText:SetText("Show Druid Innervate Cooldowns")
	WhatsUp_ShowDruidRebirthCooldownsCheckboxText:SetText("Show Druid Rebirth Cooldowns")
	WhatsUp_AnnounceSpellcastToRaidChatCheckboxText:SetText("Announce Spellcast To Raid Chat")
	WhatsUp_ShowHideUponEnteringLeavingGroupCheckboxText:SetText("Show/Hide Upon Entering/Leaving Group")
	
	WhatsUp_MasterLooterCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowSoulstonedPlayersCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowWarlockSoulstoneCooldownsCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowFearWardedPlayersCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowPriestFearWardCooldownsCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowDIdPlayersCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowPaladinDivineInterventionCooldownsCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowDruidInnervateCooldownsCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowDruidRebirthCooldownsCheckboxText:SetTextColor(1,1,1)
	WhatsUp_AnnounceSpellcastToRaidChatCheckboxText:SetTextColor(1,1,1)
	WhatsUp_ShowHideUponEnteringLeavingGroupCheckboxText:SetTextColor(1,1,1)
	WhatsUp_OpacitySlider:SetValue(OptionsList["BackgroundOpacity"])
	
	WhatsUp_MasterLooterCheckbox:SetChecked(OptionsList["ShowMasterLooter"])
	WhatsUp_ShowSoulstonedPlayersCheckbox:SetChecked(OptionsList["ShowSoulstonedPlayers"])
	WhatsUp_ShowWarlockSoulstoneCooldownsCheckbox:SetChecked(OptionsList["ShowWarlockSoulstoneCooldowns"])
	WhatsUp_ShowFearWardedPlayersCheckbox:SetChecked(OptionsList["ShowFearWardedPlayers"])
	WhatsUp_ShowPriestFearWardCooldownsCheckbox:SetChecked(OptionsList["ShowPriestFearWardCooldowns"])
	WhatsUp_ShowDIdPlayersCheckbox:SetChecked(OptionsList["ShowDIdPlayers"])
	WhatsUp_ShowPaladinDivineInterventionCooldownsCheckbox:SetChecked(OptionsList["ShowPaladinDivineInterventionCooldowns"])
	WhatsUp_ShowDruidInnervateCooldownsCheckbox:SetChecked(OptionsList["ShowDruidInnervateCooldowns"])
	WhatsUp_ShowDruidRebirthCooldownsCheckbox:SetChecked(OptionsList["ShowDruidRebirthCooldowns"])
	WhatsUp_AnnounceSpellcastToRaidChatCheckbox:SetChecked(OptionsList["AnnounceSpellcastToRaidChat"])
	WhatsUp_ShowHideUponEnteringLeavingGroupCheckbox:SetChecked(OptionsList["ShowHideUponEnteringLeavingGroup"])
	
	WhatsUp_OpacitySliderLow:SetText("Light")
	WhatsUp_OpacitySliderHigh:SetText("Dark")
	WhatsUp_OpacitySliderText:SetText("Background Opacity")
	
	WhatsUp_OptionsFrame:Show()

end

------------------------------------------------------------------------------------

function WhatsUp:OnCloseOptionsButtonClick()

	-- Set the options in the table after closing the options form
	
	OptionsList["ShowMasterLooter"] = WhatsUp_MasterLooterCheckbox:GetChecked()
	OptionsList["ShowSoulstonedPlayers"] = WhatsUp_ShowSoulstonedPlayersCheckbox:GetChecked()
	OptionsList["ShowWarlockSoulstoneCooldowns"] = WhatsUp_ShowWarlockSoulstoneCooldownsCheckbox:GetChecked()
	OptionsList["ShowFearWardedPlayers"] = WhatsUp_ShowFearWardedPlayersCheckbox:GetChecked()
	OptionsList["ShowPriestFearWardCooldowns"] = WhatsUp_ShowPriestFearWardCooldownsCheckbox:GetChecked()
	OptionsList["ShowDIdPlayers"] = WhatsUp_ShowDIdPlayersCheckbox:GetChecked()
	OptionsList["ShowPaladinDivineInterventionCooldowns"] = WhatsUp_ShowPaladinDivineInterventionCooldownsCheckbox:GetChecked()
	OptionsList["ShowDruidInnervateCooldowns"] = WhatsUp_ShowDruidInnervateCooldownsCheckbox:GetChecked()
	OptionsList["ShowDruidRebirthCooldowns"] = WhatsUp_ShowDruidRebirthCooldownsCheckbox:GetChecked()
	OptionsList["AnnounceSpellcastToRaidChat"] = WhatsUp_AnnounceSpellcastToRaidChatCheckbox:GetChecked()
	OptionsList["ShowHideUponEnteringLeavingGroup"] = WhatsUp_ShowHideUponEnteringLeavingGroupCheckbox:GetChecked()
	OptionsList["BackgroundOpacity"] = WhatsUp_OpacitySlider:GetValue()
	
	WhatsUp_OptionsFrame:Hide()
	WhatsUp:SetMaxSections()
	WhatsUp:RedrawPlayerList()

end

------------------------------------------------------------------------------------

function WhatsUp:OnOpacitySliderValueChanged(Self, Value, UserInput)

	OptionsList["BackgroundOpacity"] = WhatsUp_OpacitySlider:GetValue()
	WhatsUp:SetBackgroundOpacity()

end

------------------------------------------------------------------------------------

function WhatsUp:SetMaxSections()

	MaxSections = 0

	if OptionsList["ShowMasterLooter"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowSoulstonedPlayers"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowWarlockSoulstoneCooldowns"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowFearWardedPlayers"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowPriestFearWardCooldowns"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowDIdPlayers"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowPaladinDivineInterventionCooldowns"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowDruidInnervateCooldowns"] then
		MaxSections = MaxSections + 1
	end
	
	if OptionsList["ShowDruidRebirthCooldowns"] then
		MaxSections = MaxSections + 1
	end

	SectionToShowFirst = 1

end

------------------------------------------------------------------------------------

RegisterSuccess = C_ChatInfo.RegisterAddonMessagePrefix(CAddonID)
WhatsUp_EventFrame:RegisterEvent("GROUP_ROSTER_UPDATE")
WhatsUp_EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
WhatsUp_EventFrame:RegisterEvent("PLAYER_LOGIN")
WhatsUp_EventFrame:RegisterEvent("PLAYER_LOGOUT")
WhatsUp_EventFrame:RegisterEvent("PLAYER_DEAD")
WhatsUp_EventFrame:RegisterEvent("PLAYER_UNGHOST")
WhatsUp_EventFrame:RegisterEvent("PLAYER_ALIVE")
WhatsUp_EventFrame:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
WhatsUp_EventFrame:RegisterEvent("UNIT_INVENTORY_CHANGED")
WhatsUp_EventFrame:RegisterEvent("CHAT_MSG_ADDON")
WhatsUp_EventFrame:SetScript("OnEvent", WhatsUp_EventHandler)
C_Timer.NewTicker(1, WhatsUp_OnSecond)

